# -*- Mode:Python; coding: utf-8 -*-

# Having a go at the constraint-based inference algorithm described
#  by Pottier and Remy in ATTPL, chapter 10: "The Essence of ML Type Inference".
#
# Another great reference is a somewhat simplified presentation of the
#  same material, but (thankfully) with some context, by Pottier: "A
#  modern eye on ML type inference - Old techniques and recent
#  developments", available from his home page:
#  http://cristal.inria.fr/~fpottier/
#
# NOTE: this is a work in progress.
# http://www.nightmare.com/rushing/irken
#
# I'm trying *for now* to make this as close to the formal spec as possible,
# in order to understand the algorithm, which makes this Not Very Efficient.
# After that, I plan to do the usual union-find/path-compression tvar stuff,
# and try to grok the 'rank' attribute as described in the chapter.

# TODO: s-letall, letrec, kind-checking

import nodes

from pprint import pprint
import sys
is_a = isinstance

import pdb
trace = pdb.set_trace

# the simply typed lambda calculus:
#   e ::= x | \x.e | e e
# 
# expressions:
# x    : <varref>
# \x.e : <function>
# e e  : <application>

# types
# t ::= a | (arrow t t)
# (where a = <tvar>)

# constraints:
# C ::= (equals t t) | (and C C) | (exists a C)

# types

tvar_counter = 0

def fresh():
    global tvar_counter
    result = tvar_counter
    tvar_counter += 1
    return result

class _type:
    pass

class t_base (_type):
    def __cmp__ (self, other):
        return cmp (self.__class__, other.__class__)

class t_int (t_base):
    def __repr__ (self):
        return 'int'

class t_char (t_base):
    def __repr__ (self):
        return 'char'

class t_str (t_base):
    def __repr__ (self):
        return 'str'

class t_var (_type):
    letters = 'abcdefghijklmnopqrstuvwxyz'
    def __init__ (self):
        self.id = fresh()
    def __repr__ (self):
        if self.id < len(self.letters):
            return self.letters[self.id]
        else:
            return 'T%d' % (self.id - len(self.letters))

class t_predicate (_type):
    def __init__ (self, name, args):
        self.name = name
        self.args = tuple (args)
    def __repr__ (self):
        # special case
        if self.name == 'arrow':
            return '%r->%r' % self.args
        else:
            return '%s%r' % (self.name, self.args)

def is_pred (t, *p):
    # is this a predicate from the set <p>?
    return is_a (t, t_predicate) and t.name in p

def arrow (t0, t1):
    return t_predicate ('arrow', (t0, t1))
    
# row types
def product (row):
    # a.k.a. 'Π'
    # XXX kind-check that args[0] is a row?
    return t_predicate ('product', (row,))

def sum (row):
    # a.k.a. 'Σ'
    return t_predicate ('sum', (row,))

def rdefault (arg):
    # a.k.a. 'δ'
    return t_predicate ('rdefault', (arg,))

def rlabel (name, type, rest):
    return t_predicate ('rlabel', (name, type, rest))

# constraints

def constraint_repr (kind, args):
    if args:
        return '[%s %s]' % (kind, ' '.join ([repr(x) for x in args]))
    else:
        return '%s' % (kind,)

class constraint:
    kind = 'abstract'
    args = ()

    def __repr__ (self):
        return constraint_repr (self.kind, self.args)

class c_true (constraint):
    kind = 'true'
    args = ()

class c_false (constraint):
    kind = 'true'
    args = ()

class c_equals (constraint):
    kind = 'equals'
    def __init__ (self, *args):
        # this class is used for both simple binary equality, but also
        # as a 'multi-equation' for unification, hence the *args
        self.args = args
        if len(args) == 0:
            raise ValueError
    def __repr__ (self):
        return constraint_repr ('=', self.args)
    def split (self):
        # split into tvars and types
        # a multi-equation in standard form will have at most one non-variable.
        tv, ty = [], []
        for e in self.args:
            if is_a (e, t_var):
                tv.append (e)
            else:
                ty.append (e)
        return tv, ty

class c_and (constraint):
    kind = 'and'
    def __init__ (self, c0, c1):
        self.args = (c0, c1)
        
class c_exists (constraint):
    kind = 'exists'
    def __init__ (self, vars, sub):
        self.args = (vars, sub)
        self.vars = vars
        self.sub = sub

class c_is (constraint):
    # <x> has type <t> iff <t> is an instance of the type scheme associated with <x>
    kind = 'is'
    def __init__ (self, x, t):
        self.args = (x, t)
        self.x = x
        self.t = t

class c_let (constraint):
    kind = 'let'
    def __init__ (self, formal, init, body):
        self.args = (formal, init, body)
        self.formal = formal
        self.init = init # this is poorly named, should be 'type' or 'scheme'
        self.body = body

class c_forall (constraint):
    kind = 'forall'
    def __init__ (self, vars, constraint, type):
        self.args = (vars, constraint, type)
        self.vars = vars
        self.constraint = constraint
        self.type = type

# stack frames
class frame:
    kind = 'abstract'
    args = ()

class s_empty (frame):
    kind = 'empty'

empty = s_empty()

class s_and (frame):
    def __init__ (self, c):
        self.args = (c,)
        self.constraint = c

class s_exists (frame):
    def __init__ (self, vars):
        self.args = (vars,)
        self.vars = vars

class s_let (frame):
    def __init__ (self, formal, vars, type, body):
        self.args = (formal, vars, type, body)
        self.formal = formal
        self.vars = vars
        self.type = type
        self.body = body

class s_env (frame):
    def __init__ (self, formal, type):
        self.args = (formal, type)
        self.formal = formal
        self.type = type

# this is a two-phase algorithm
# 1) constraint generation
# 2) constraint solving

class constraint_generator:

    def go (self, exp):
        t = t_var()
        return self.gen (exp, t), t

    def gen (self, exp, t):
        if is_a (exp, nodes.varref):
            return c_is (exp.name, t)
        elif is_a (exp, nodes.function):
            assert (len (exp.formals) == 1)
            a1 = t_var()
            a2 = t_var()
            sub0 = self.gen (exp.body, a2)
            let0 = c_let (exp.formals[0].name, a1, sub0)
            # XXX: in ATTPL, this is a c_supertype relation
            sub1 = c_equals (t, arrow (a1, a2))
            return c_exists ((a1, a2), c_and (let0, sub1))
        elif is_a (exp, nodes.application):
            e1 = exp.rator
            # simple lambda calculus only!
            assert (len (exp.rands) == 1)
            e2 = exp.rands[0]
            a = t_var()
            sub0 = self.gen (e1, arrow (a, t))
            sub1 = self.gen (e2, a)
            return c_exists ((a,), c_and (sub0, sub1))
        elif is_a (exp, nodes.let):
            x = t_var()
            assert (len(exp.names) == 1)
            init0 = self.gen (exp.inits[0], x)
            body0 = self.gen (exp.body, t)
            return c_let (exp.names[0].name, c_forall ((x,), init0, x), body0)
        elif is_a (exp, nodes.literal):
            if exp.kind == 'int':
                return c_equals (t, t_int())
            elif exp.kind == 'char':
                return c_equals (t, t_char())
            else:
                raise ValueError ("unsupported literal type")
        else:
            raise ValueError

class UnboundVariable (Exception):
    pass

class Continue (Exception):
    pass

class TypeError (Exception):
    pass

# I believe the 'union-find' data-structure/algorithm is another way
#   of describing HM's use of type variable lookup with path
#   compression.

class solver:

    def __init__ (self, step=True):
        self.step = step
        self.prim_env = self.make_prim_env()

    def make_prim_env (self):
        # build type schemes for builtins/primitives
        int = t_int()
        char = t_char()
        e = {}
        e['%+'] = arrow (int, arrow (int, int))
        e['%-'] = arrow (int, arrow (int, int))
        # XXX consider generating the tvars fresh every time
        x = t_var()
        y = t_var()
        e['%make-pair'] = c_forall ((x,y), c_true(), arrow (x, arrow (y, t_predicate ('pair', (x, y)))))
        e['%pair/first'] = c_forall ((x,y), c_true(), arrow (t_predicate ('pair', (x, y)), x))
        e['%pair/second'] = c_forall ((x,y), c_true(), arrow (t_predicate ('pair', (x, y)), y))
        e['%char->int'] = arrow (char, int)
        e['%int->char'] = arrow (int, char)
        # rows
        e['%make-row'] = c_forall ((x,), c_true(), arrow (x, product (rdefault (x))))
        # label-related row types are in the 'lookup_special' method
        return e

    def dprint (self, msg):
        if self.step:
            sys.stderr.write (msg)
            sys.stderr.write ('\n')

    def solve (self, c):

        self.dprint ('\nHit <return> at each pause (or "t<return>" to enter the debugger)')

        pvars = []
        exists = []
        # ensure there are always two items on the stack
        s = [empty, empty]
        u = []
        c = c
        orig_c = c

        # for when all of U is wrapped in an <exists>
        self.u_exists = []

        def push (x):
            s.append (x)

        def pop ():
            s.pop()

        while 1:

            if self.step:
                print 'S:',
                self.pprint_stack (s)
                print 'U:', u
                print 'C:', c
                print 'exists:', exists

            # the top two elements of the stack
            sy, sz = s[-2], s[-1]

            # --- unification ---

            while 1:
                unified = False
                u2 = []

                u = self.try_name_1 (u)
                u = self.try_fuse (u)
                u = self.try_decompose (u)
                
                for eq in u:
                    if is_a (eq, _type):
                        self.dprint ('s-single')
                        unified = True
                    elif is_a (eq, c_true):
                        self.dprint ('s-true')
                        unified = True
                    elif is_a (eq, c_exists):
                        # XXX alpha rename this guy and attach the variables at top level
                        self.dprint ('s-exand')
                        u2.append (eq.sub)
                        unified = True
                        # this will get caught by s-ex-1
                        self.u_exists.extend (eq.vars)
                    elif is_a (eq, c_equals):
                        if len(eq.args) == 1:
                            self.dprint ('s-single')
                            unified = True
                        else:
                            tv, ty = eq.split()
                            tvs = set(tv) # remove-duplicates
                            if len(tvs) != len(tv):
                                self.dprint ('s-stutter')
                                tv = tvs
                                unified = True
                            if len(ty) > 1:
                                # FX=FT=ε => X=T ^ FX=ε
                                print 'this probably should not happen here'
                                trace()
                                pass
                            u2.append (c_equals (* (list(tv) + ty)))
                    else:
                        raise ValueError ("unknown object in U")
                if not unified:
                    break
                else:
                    u = u2
                    self.dprint ('U: %r' % u)

            if self.step:
                print '-----------------------------'
                if raw_input().startswith ('t'):
                    trace()

            # --- solver ---            

            # XXX everywhere here, alpha-conversion to avoid capture (i.e., all the X!=ftv(e) rules)
            if self.u_exists:
                self.dprint ('s-ex-1')
                push (s_exists (tuple (self.u_exists)))
                self.u_exists = []
            elif is_a (sy, s_and) and is_a (sz, s_exists):
                self.dprint ('s-ex-2')
                # swap 'em
                pop(); pop()
                push (sz); push (sy)
            elif is_a (sy, s_let) and is_a (sz, s_exists):
                self.dprint ('s-ex-3')
                pop(); pop()
                push (s_let (sy.formal, sy.vars + sz.vars, sy.type, sy.body))
            elif is_a (sy, s_env) and is_a (sz, s_exists):
                self.dprint ('s-ex-4')
                pop(); pop()
                push (sz)
                push (sy)
            elif is_a (sy, s_exists) and is_a (sz, s_exists):
                # XXX hack?
                self.dprint ('s-ex-5')
                pop(); pop();
                push (s_exists (sy.vars + sz.vars))
            elif is_a (sz, s_exists):
                self.dprint ('s-record-ex')
                exists.extend (sz.vars)
                pop()
            elif is_a (c, c_equals):
                self.dprint ('s-solve-eq')
                u.append (c)
                c = c_true()
            elif is_a (c, c_is) and is_a (c.x, str):
                self.dprint ('s-solve-id')
                scheme, type = self.lookup (c.x, s), c.t
                # assert that scheme.type is a tvar
                # "Recall that if σ is of the form ∀X0..XN[U].X
                # where X0..XN#ftv(T), then c_is(σ, T) stands for ∃X0..XN.(U ^ X=T)."
                self.dprint ('scheme= %r' % scheme)
                self.dprint ('type=%r' % type)
                if is_a (scheme, c_forall):
                    if not scheme.vars and is_a (scheme.constraint, c_true):
                        c = c_equals (scheme.type, type)
                    else:
                        c = c_exists (scheme.vars, c_and (scheme.constraint, c_equals (scheme.type, type)))
                else:
                    c = c_equals (scheme, type)
            elif is_a (c, c_and):
                self.dprint ('s-solve-and')
                push (s_and (c.args[1]))
                c = c.args[0]
            elif is_a (c, c_exists):
                self.dprint ('s-solve-ex')
                push (s_exists (c.vars))
                c = c.sub
            elif is_a (c, c_let):
                self.dprint ('s-solve-let')
                if is_a (c.init, c_forall):
                    push (s_let (c.formal, c.init.vars, c.init.type, c.body))
                    c = c.init.constraint
                else:
                    # let x: T in C == let x: ∀∅[true].T in C
                    push (s_let (c.formal, (), c.init, c.body))
                    c = c_true()
            elif is_a (c, c_true):
                if is_a (sz, s_and):
                    self.dprint ('s-pop-and')
                    pop()
                    c = sz.constraint
                elif is_a (sz, s_let) and not is_a (sz.type, t_var):
                    self.dprint ('s-name-2')
                    x = t_var()
                    pop()
                    push (s_let (sz.formal, sz.vars + (x,), x, sz.body))
                    u.append (c_equals (x, sz.type))
                elif is_a (sz, s_let):
                    # s-compress, s-unname, s-letall, s-pop-let
                    # these rules all require an examination of the variables
                    #  bound by this <let>.
                    try:
                        ftvu = self.ftvu (u)
                        self.dprint ('ftvu= %r' % ftvu)
                        for var in sz.vars:
                            if ftvu.count (var) == 0 and sz.type is not var:
                                # completely unused variable
                                self.dprint ('s-unname-2 (Y=%r)' % (var,))
                                pop()
                                vars = tuple ([x for x in sz.vars if x is not var])
                                push (s_let (sz.formal, vars, sz.type, sz.body))
                                raise Continue
                            for e in u:
                                tv, ty = e.split()
                                if var in tv and len(tv) >= 2:
                                    # for s-compress, we want to know if there's a multi-equation
                                    # of the form Y=Z=ε (where Y is bound by this <let>)
                                    self.dprint ('s-compress')
                                    # subst will leave this one in the form Z=Z=ε, which
                                    #   should be caught by s-stutter in the unifier.
                                    Y = var
                                    if tv[0] is not var:
                                        Z = tv[0]
                                    else:
                                        Z = tv[1]
                                    self.dprint ('Y=%r' % Y)
                                    self.dprint ('Z=%r' % Z)
                                    u = [ self.subst (Y, Z, x) for x in u ]
                                    u.append (Y)
                                    if sz.type == var:
                                        pop()
                                        push (s_let (sz.formal, sz.vars, Z, sz.body))
                                    raise Continue
                                elif (var in tv
                                      and len (self.find_tvar (u, var)) == 1
                                      and ftvu.count (var) == 1
                                      and sz.type is not var
                                      ) :
                                    self.dprint ('s-unname')
                                    self.dprint (' Y=%r' % var)
                                    vars = tuple([x for x in sz.vars if x is not var])
                                    pop()
                                    # remove Y from the <let>
                                    push (s_let (sz.formal, vars, sz.type, sz.body))
                                    # remove Y from <u> - since we know it's only in this
                                    #  one multi-equation, I'm gonna cheat and just mutate it.
                                    e.args = [x for x in e.args if x is not var]
                                    raise Continue
                        #elif s-letall:
                        # s-letall will simplify the current scheme by removing some of
                        #   the quantified tvars.
                        else:
                            # pop-let is a fall-through - after all
                            # the above conditions have been met it turns the <let>
                            # into an <env>.
                            self.dprint ('s-pop-let')
                            u1, u2 = self.split_u (sz, u)
                            self.dprint ('u1=%r' % u1)
                            self.dprint ('u2=%r' % u2)
                            pop()
                            push (s_env (sz.formal, c_forall (sz.vars, list_to_conj (u2), sz.type)))
                            u = u1
                            c = sz.body
                    except Continue:
                        # back to the top of the solver loop
                        pass
                elif is_a (sz, s_env):
                    pvars.append ((sz.formal, sz.type.type))
                    self.dprint ('s-pop-env')
                    pop()
                elif is_a (sz, s_empty):
                    self.dprint ('exists=%r' % exists)
                    self.dprint ('constraint=%r' % orig_c)
                    return pvars, u
                else:
                    raise ValueError ("unexpected")
            else:
                raise ValueError ("no rule applies")
                
    def split_u (self, sz, u):
        # u1 is made up entirely of 'old' variables - i.e., not bound by
        #  the <let> in sz.
        u1 = []
        u2 = []
        young = set (sz.vars)
        for e in u:
            flag = True
            for var in ftv (e):
                # only if all of them are old can it go in u1
                if var in young:
                    flag = False
            if flag:
                u1.append (e)
            else:
                u2.append (e)
        return u1, u2

    def find_tvar (self, u, v):
        r = []
        for eq in u:
            tv, ty = eq.split()
            if v in tv:
                r.append (eq)
        return r

    # yes, this way of managing tvars and U is very clumsy.  this code will become much
    #   simpler...

    def try_name_1 (self, u):
        u2 = []
        eqs = []
        for eq in u:
            t2 = []
            if is_a (eq, c_equals):
                for t in eq.args:
                    # looking for a predicate with any non-variable arguments
                    if is_a (t, t_predicate):
                        args2 = []
                        for arg in t.args:
                            if is_a (arg, str):
                                # XXX row labels, must be a better way
                                args2.append (arg)
                            elif not is_a (arg, t_var):
                                self.dprint ('s-name-1')
                                tv = t_var()
                                self.u_exists.append (tv)
                                args2.append (tv)
                                u2.append (c_equals (tv, arg))
                            else:
                                args2.append (arg)
                        if len(args2):
                            t2.append (t_predicate (t.name, tuple(args2)))
                        else:
                            t2.append (t)
                    else:
                        t2.append (t)
                u2.append (c_equals (*t2))
            else:
                u2.append (eq)
        return u2

    def try_fuse (self, u):
        self.dprint ('before fuse: %r' % u)
        map = {}
        u2 = []
        for eq in u:
            if is_a (eq, c_equals):
                tvs, tys = eq.split()
                did_fuse = False
                for tv in set(tvs):
                    if not map.has_key (tv.id):
                        map[tv.id] = eq
                    else:
                        eq0 = map[tv.id]
                        self.dprint ('s-fuse %r %r' % (eq0, eq))
                        if eq0 is eq:
                            raise ValueError ("I'm confused")
                        # A=B=T0 ^ B=C=T1 => A=B=C=T0=T1
                        tvs0, tys0 = eq0.split()
                        u2.remove (eq0)
                        for tv in tvs0:
                            del map[tv.id]
                        # find the set of unique tvars
                        tvs1 = set (tvs0).union (set (tvs))
                        u2.append (c_equals (*list(tvs1) + tys0 + tys))
                        did_fuse = True
                        for tv in tvs1:
                            map[tv.id] = u2[-1]
                        # take care of any other repeats in the next round
                        break
                if not did_fuse:
                    u2.append (eq)
            else:
                # true | exists
                u2.append (eq)
        self.dprint ('after fuse: %r' % u2)
        return u2

    def try_decompose (self, u):
        u2 = []
        for eq in u:
            if is_a (eq, c_equals):
                tvs, tys = eq.split()
                if len(tys) > 1:
                    # two non-variable members, a non-standard multi-equation.
                    # shouldn't be possible to collect more than two non-vars.
                    assert (len(tys) == 2)
                    ty0, ty1 = tys[:2]
                    if ty0 == ty1:
                        # a=b=int=int, etc... => a=b=int
                        u2.append (c_equals (*tvs + [ty0]))
                    elif is_pred (ty0, 'rlabel', 'rdefault') or is_pred (ty1, 'rlabel', 'rdefault'):
                        u2.extend (self.try_row_unification (ty0, ty1, tvs))
                    elif (is_a (ty0, t_predicate) and is_a (ty1, t_predicate)
                        and ty0.name == ty1.name
                        and len(ty0.args) == len(ty1.args)):
                        self.dprint ('s-decompose')
                        if ty0.name in ('rlabel', 'rdefault'):
                            trace()
                        # P(a,b,c)=P(d,e,f)=ε => a=d ^ b=e ^ c=f ^ P(a,b,c)=ε
                        for i in range (len (ty0.args)):
                            u2.append (c_equals (ty0.args[i], ty1.args[i]))
                        u2.append (c_equals (* tvs + [ty0]))
                    else:
                        self.dprint ('s-clash')
                        raise TypeError ((ty0, ty1))
                else:
                    u2.append (eq)
            else:
                u2.append (eq)
        return u2

    def try_row_unification (self, ty0, ty1, tvs):
        if is_pred (ty0, 'rlabel') and is_pred (ty1, 'rlabel'):
            if ty0.args[0] != ty1.args[0]:
                # distinct head labels
                self.dprint ('s-mutate-ll')
                # XXX be concerned about how one of these may have types
                #     and the other has variables.  do we need to check
                #     and reorder them?
                l0, t0, d0 = ty0.args
                l1, t1, d1 = ty1.args
                x = t_var()
                self.u_exists.append (x)
                return [
                    c_equals (d0, rlabel (l1, t1, x)),
                    c_equals (d1, rlabel (l0, t0, x)),
                    c_equals (*tvs + [rlabel (l0, t0, d0)])
                    ]
            else:
                # XXX this should be handled by the normal s-decompose
                l0, t0, d0 = ty0.args
                l1, t1, d1 = ty1.args
                return [
                    c_equals (t0, t1),
                    c_equals (d0, d1),
                    c_equals (*tvs + [ty0])
                    ]
        elif is_pred (ty0, 'rlabel') or is_pred (ty1, 'rlabel'):
            # only one is an rlabel
            if is_pred (ty1, 'rlabel'):
                # ensure that ty0 is the rlabel
                ty0, ty1 = ty1, ty0
            if is_pred (ty1, 'rdefault'):
                self.dprint ('s-mutate-dl')
                x = ty1.args[0]
                assert (is_a (x, t_var))
                return [
                    c_equals (x, ty0.args[1]),
                    c_equals (ty1, ty0.args[2]),
                    c_equals (*tvs + [ty1])
                    ]
            elif is_a (ty1, t_predicate):
                # some other predicate
                self.dprint ('s-mutate-gl')
                trace()
                n = len (ty1.args)
                tvars0 = [t_var() for x in ty1.args]
                tvars1 = [t_var() for x in ty1.args]
                self.u_exists.extend (tvars0)
                self.u_exists.extend (tvars1)
                l0, t0, d0 = ty0.args
                g = ty1.name
                return [
                    c_equals (t_predicate (g, tvars0), t0),
                    c_equals (t_predicate (g, tvars1), d0),
                    ] + [ c_equals (ty1.args[i], rlabel (l0, tvars0[i], tvars1[i])) for i in range (n) ] + [
                    c_equals (*tvs + [ty1])
                    ]
            else:
                self.dprint ('s-clash')
                raise TypeError ((ty0, ty1))
        elif is_pred (ty0, 'rdefault',) or is_pred (ty1, 'rdefault'):
            if is_pred (ty1, 'rdefault'):
                # ensure that ty0 is the rdefault/δ
                ty0, ty1 = ty1, ty0
            if is_a (ty1, t_predicate):
                # some other predicate
                self.dprint ('s-mutate-gd')
                n = len (ty1.args)
                g = ty1.name
                tvars = [ t_var() for x in ty1.args ]
                self.u_exists.extend (tvars)
                return [
                    c_equals (ty0.args[0], t_predicate (g, tvars))
                    ] + [ c_equals (ty1.args[i], rdefault (tvars[i])) for i in range (n) ] + [
                    c_equals (*tvs + [ty0])
                    ]
            else:
                self.dprint ('s-clash')
                raise TypeError ((ty0, ty1))
        else:
            self.dprint ('s-clash')
            raise TypeError ((ty0, ty1))

    def lookup (self, x, s):
        # XXX side conditions
        for f in s:
            if is_a (f, s_and):
                continue
            elif is_a (f, s_exists):
                continue
            elif is_a (f, s_let):
                if f.formal == x:
                    raise ValueError ("shouldn't happen?")
                continue
            elif is_a (f, s_env):
                if f.formal != x:
                    continue
                else:
                    return f.type
        else:
            return self.lookup_special_names (x)

    def lookup_special_names (self, name):
        # we need this hack for row-related label lookups, since the label
        #  is part of the name.
        if self.prim_env.has_key (name):
            return self.prim_env[name]
        elif name.startswith ('%rextend/'):
            what, label = name.split ('/')
            x, y, z = t_var(), t_var(), t_var()
            # ∀XYZ.Π(l:X;Y) → Z → Π(l:Z;Y)
            return c_forall (
                (x,y,z),
                c_true(),
                arrow (product (rlabel (label, x, y)),
                       arrow (z, product (rlabel (label, z, y))))
                )
        elif name.startswith ('%raccess/'):
            what, label = name.split ('/')
            x, y = t_var(), t_var()
            # ∀XY.Π(l:X;Y) → X
            return c_forall ((x, y), c_true(), arrow (product (rlabel (label, x, y)), x))
        else:
            raise UnboundVariable (name)

    # need a 'map_constraint' that will map a function over any constraint.
    def subst (self, v0, v1, e):
        # replace all <v0> with <v1>
        if self.step:
            print ' -- subst before:', e
        def f (e):
            if is_a (e, t_var):
                if e.id == v0.id:
                    return v1
                else:
                    return e
            elif is_a (e, t_predicate):
                return t_predicate (e.name, [f(x) for x in e.args])
            if is_a (e, c_equals):
                return c_equals (*[f(x) for x in e.args])
            elif is_a (e, c_and):
                return c_and (*[f(x) for x in e.args])
            elif is_a (e, t_base):
                return e
            elif is_a (e, str):
                # XXX fixme - this happens with row labels, might need their own predicate?
                return e
            else:
                raise ValueError ("unknown expression in subst")
        r = f (e)
        if self.step:
            print ' -- subst after :', r
        return r

    def ftvu (self, u):
        r = []
        for e in u:
            r.extend (ftv (e))
        return r

    def pprint_stack (self, s):

        W = sys.stdout.write

        W ('\n')
        n = len(s)
        # the 2 is for the two <empty> sentinels
        for i in range (2,n):
            W ('%2d: ' % (i-2,))
            si = s[i]
            if is_a (si, s_empty):
                W ('[]')
            elif is_a (si, s_and):
                W ('[] ^ %s' % si.args)
            elif is_a (si, s_exists):
                W ('exists%r.[]' % si.args)
            elif is_a (si, s_let):
                W ('let %s: forall%r[[]].%r in %r' % si.args)
            elif is_a (si, s_env):
                W ('env %s: %r in []' % si.args)
            else:
                raise NotImplementedError
            W ('\n')


def ftv (e):
    r = []
    def recur (bound, e):
        if is_a (e, c_equals):
            for arg in e.args:
                recur (bound, arg)
        elif is_a (e, c_and):
            for arg in e.args:
                recur (bound, arg)
        elif is_a (e, c_exists):
            recur (bound.union (e.vars), e.sub)
        elif is_a (e, c_is):
            recur (bound, e.t)
        elif is_a (e, c_let):
            recur (bound, e.init)
            recur (bound, e.body)
        elif is_a (e, c_forall):
            bound = bound.union (e.vars)
            recur (bound, e.constraint)
            recur (bound, e.type)
        elif is_a (e, t_var):
            r.append (e)
        elif is_a (e, t_predicate):
            for arg in e.args:
                recur (bound, arg)
        elif is_a (e, t_base):
            pass
        elif is_a (e, c_true):
            pass
        elif is_a (e, c_false):
            pass
        elif is_a (e, str):
            # XXX fixme - this happens with row labels, might need their own predicate?
            pass
        else:
            raise NotImplementedError
    recur (set(), e)
    return r

def list_to_conj (l):
    if len(l) == 0:
        return c_true()
    elif len(l) == 1:
        return l[0]
    else:
        r = l[0]
        for x in l[1:]:
            r = c_and (r, x)
        return r

def map_type (t, f):
    if is_a (t, t_var):
        return f (t)
    elif is_a (t, t_predicate):
        return t_predicate (t.name, [f(x) for x in t.args])
    elif is_a (t, t_base):
        return f (t)
    elif is_a (t, str):
        return t
    else:
        raise ValueError ("unknown type object")

def print_solution (pvars, u, top_tv):

    print pvars
    print u
    print top_tv

    def lookup (v):
        for eq in u:
            tv, ty = eq.split()
            if v in tv:
                if len(ty):
                    return map_type (ty[0], lookup)
                else:
                    trace()
        else:
            return v

    for pvar, tvar in pvars:
        print '%r: %r' % (pvar, lookup (tvar))

    print 'program: %r' % lookup (top_tv)

def read_string (s):
    import cStringIO
    import lisp_reader
    sf = cStringIO.StringIO (s)
    r = lisp_reader.reader (sf)
    return r.read()

def test (s, step=True):
    import nodes
    from pprint import pprint as pp
    global tvar_counter
    tvar_counter = 0
    # wrap everything in a top-level <let>
    # if we omit this, s-compress won't work on top-level exists
    print 'expression=', s
    s = "(let ((top %s)) top)" % s
    exp = read_string (s)
    w = nodes.walker()
    exp2 = w.go (exp)
    # alpha conversion
    nodes.rename_variables (exp2)
    cg = constraint_generator()
    c, top_tv = cg.go (exp2)
    print 'constraint=', c
    m, u = solver(step).solve(c)
    print_solution (m, u, top_tv)

tests = [
    "5",
    "(lambda (x) 3)",
    "(lambda (x) x)",
    "(let ((f (lambda (x) x))) (f 5))",
    "(lambda (x) (lambda (y) x))",
    "(let ((f (lambda (x) x))) f)",
    "((lambda (x) x) (lambda (x) x))",
    "(%+ 3)",
    "((%+ 3) 4)",
    "((%- ((%+ 3) 4)) 2)",
    "((%make-pair 3) 4)",
    "((%make-pair 3) #\\A)",
    "(%pair/first ((%make-pair 3) #\\A))",
    "(%pair/second ((%make-pair 3) #\\A))",
    "(%char->int #\\A)",
    "(%int->char 65)",
    # this causes a cycle, breaks print-solution
    #"(let ((f (lambda (x) 3))) (%int->char (f f)))",
    # row tests
    # make a default row - maps all symbols to a function: δ(a->a)
    "(%make-row (lambda (x) x))",
    # lookup via a random field name on a row of type δ(int)
    "(%raccess/field (%make-row 0))",
    # extend a row: (label:char;δ(int))
    "((%rextend/label (%make-row 0)) #\\A)",
    # dereference via the defined label
    "(%raccess/label ((%rextend/label (%make-row 0)) #\\A))",
    # dereference via an undefined label
    "(%raccess/fnord ((%rextend/label (%make-row 0)) #\\A))",
    # extend a row twice: (other:a->a; label:char; δ(int))
    "((%rextend/other ((%rextend/label (%make-row 0)) #\\A))(lambda (x) x))",
    # deref one defined label
    "(%raccess/label ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # deref the other
    "(%raccess/other ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # deref an undefined label
    "(%raccess/fnord ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x)))",
    # build an extended record, store it in a variable, then deref to pull the identity function out and use it.
"""(let ((rec0 ((%rextend/other ((%rextend/label (%make-row 0)) #\\A)) (lambda (x) x))))
     (let ((f0 (%raccess/other rec0)))
       (f0 9)))""",
    ]

if __name__ == '__main__':
    if '-t' in sys.argv:
        for t in tests:
            test (t, step=False)
    elif '-p' in sys.argv:
        import profile
        profile.run ("test (tests[-1], step=False)")
    elif '-i' in sys.argv:
        # interactive
        while 1:
            sys.stdout.write ('> ')
            line = raw_input()
            if not line:
                break
            else:
                #test (line, step=False)
                test (line, step=True)
    else:
        test ("(lambda (x) x)")
